"""Implementation of utility functions.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import sys
import re


def get_sbml_ids(df, sbo_id):
    """Get SBML IDs with specified sboterm-id.

    if sbotermid does not exist, return all IDs

    :param df: SBML component from which to extract IDs
    :type df: pandas DataFrame (component from sbmlxdf.to_df())
    :param sbo_id: value of sboterm to be queried, e.g. 260
    :type sbo_id: int
    :returns: component ids
    :rtype: list of str
    """
    if 'sboterm' in df:
        ids = (df[df['sboterm'] == 'SBO:{:07d}'.format(sbo_id)]).index.to_list()
    else:
        ids = df.index.to_list()
    return ids


def extract_params(record):
    """Extract parameters from a record.

    A single record consists of comma separated key-value pairs.
    Example: 'key1=val1, key2=val2, ...' is converted to
    {key1: val1, key2: val2, ...}

    :param record: key '=' value pairs separated by ","
    :type record: str
    :returns: key-values pairs extracted from record
    :rtype: dict
    """
    params = {}
    for kv_pair in record_generator(record, sep=','):
        if '=' in kv_pair:
            k, v = kv_pair.split('=')
            params[k.strip()] = v.strip()
    return params


def sympy_protect(name):
    """Modify variable names that conflict with sympy.

    see: https://docs.sympy.org/latest/gotchas.html#variables

    :param name: variable name
    :type name: str
    :returns: variable name not conflicting with sympy
    :rtype: str
    """
    if name in {'I', 'E', 'S', 'N', 'C', 'O', 'Q'}:
        name += '_sympy_protect'
    return name


def record_generator(records_str, sep=';'):
    """Generator to extract individual records from a string of records.

    :param records_str: containing records separated by sep
    :type records_str: str
    :param sep: seperator used to separate records
    :type sep: str (default: ';')
    :returns: key-values pairs extracted from record
    :rtype: dict
    """
    if type(records_str) == str:
        for record in records_str.split(sep):
            if len(record.strip()) > 0:
                yield record.strip()


def get_local_params(reaction):
    """Retrieve local parameters of a reaction and their values.

    :param reaction: reaction record from SBML reactions component
    :type reaction: pandas Series
    :returns: parameter id with parameter type, sympy_id and parameter value
    :rtype: dict (key: parameter id, val: value, units, sympy-portected id
    """
    local_params = {}
    if 'localParams' in reaction:
        for lp in record_generator(reaction['localParams']):
            params = extract_params(lp)
            local_params[params['id']] = float(params['value'])
    return local_params


def get_species_stoic(srefs):
    """Retrieve species and stoichiometry from reaction srefs.

    :param srefs: species reference records
    :type srefs: string
    :returns: species id with corresponding stoichiometry
    :rtype: dict (key: species id, val: stoichiometry (1.0 by default, e.g. for modifiers)
    """
    species = {}
    for sref in record_generator(srefs):
        params = extract_params(sref)
        species[params['species']] = float(params.get('stoic', '1.0'))
    return species


def get_species_refs(reaction):
    """Retrieve species references used in a reaction with their type.

    :param reaction: reaction record from SBML reactions component
    :type reaction: pandas Series
    :returns: species id with concentration set to 1.0 (used for enzymes)
    :rtype: dict (key: species id, val: 1.0, type, sympy-portected id
    """
    species_refs = {}
    for sref_type in ['reactants', 'products', 'modifiers']:
        if sref_type in reaction:
            for sref in record_generator(reaction[sref_type]):
                params = extract_params(sref)
                sid = params['species']
                species_refs[sid] = [1.0, sref_type, sympy_protect(sid)]
    return species_refs


# updated
def expand_kineticlaw(rate, fdefs):
    """Expand a kinetic law by inlining lambda functions

    :param rate: kinetic law
    :type rate: str
    :param fdefs: function definitions used in the model
    :type fdefs: dict of Functions
    :returns: updated kinetic law with lambda function inlined
    :rtype: str
    """
    for fid, fdef in fdefs.items():
        m = re.search(r'.*\b' + fid + r'\((.*)\)', rate)
        if m:
            calling_params = [kw.strip() for kw in m.group(1).split(',')]
            lambda_def, lambda_params = fdef.get_function()
            for i in range(len(lambda_params)):
                lambda_def = re.sub(r'\b' + lambda_params[i] + r'\b', calling_params[i], lambda_def)
            prefix_pos = re.search(fid, rate).start(0)
            rate = (rate[:prefix_pos] + ' (' + lambda_def + rate[m.end(1):])
    return rate


# display progress bar
def progress_bar(step, steps):
    sys.stdout.write('\r')
    progress = (step+1)/steps
    bar = int(progress*20)
    sys.stdout.write('[{}{}] {:3.0f}%'.format('=' * bar, ' ' * (20 - bar), progress*100))
    sys.stdout.flush()
